import typing
import asyncio
from gc import collect
from .http import AsyncRequester
from ..classes import BasePilmoji
from ..helpers import get_nodes
from aiohttp import ClientSession
from PIL import ImageFont, Image, ImageDraw


__all__ = [
    'AsyncPilmoji'
]


class AsyncPilmoji(BasePilmoji):
    """
    The synchronous emoji renderer.
    """
    def __init__(
            self,
            image: Image.Image,
            *,
            session: typing.Optional[ClientSession] = None,
            loop: typing.Optional[asyncio.AbstractEventLoop] = None,
            use_microsoft_emoji: bool = False,
            render_discord_emoji: bool = True
    ):
        if not isinstance(image, Image.Image):
            raise TypeError(f'Image must be of type Image, got {type(image).__name__!r} instead.')

        self.render_discord_emoji: bool = render_discord_emoji
        self.http: AsyncRequester = AsyncRequester(session=session, loop=loop, _microsoft=use_microsoft_emoji)
        self.image: Image.Image = image
        self.draw = ImageDraw.Draw(image)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def close(self):
        """
        Closes the requester and collects garbage.
        """
        await self.http.close()
        collect()
        del self

    async def text(self,
                   xy: typing.Tuple[int, int],
                   text: str,
                   fill=None,
                   font=None,
                   anchor=None,
                   spacing=4,
                   align="left",
                   direction=None,
                   features=None,
                   language=None,
                   stroke_width=0,
                   stroke_fill=None,
                   embedded_color=False,
                   emoji_size_factor=1,
                   emoji_position_offset=(0, 0),
                   *args, **kwargs) -> None:
        """
        Draws text with emoji rendering.
        Multiline text is supported.

        This function's signature is the exact same as PIL's, with a little bit of type-hinting.
        """
        if not font:
            font = ImageFont.load_default()

        args = (fill, font, anchor, spacing, align, direction,
                features, language, stroke_width, stroke_fill, embedded_color, *args)

        x, y = xy
        original_x = x
        lines = text.split('\n')
        nodes = get_nodes(lines)

        for line in nodes:
            x = original_x
            for node in line:
                content = node['content']
                width, height = font.getsize(content)
                if node['type'] == 'text':
                    self.draw.text((x, y), content, *args, **kwargs)
                else:
                    if node['type'] == 'twemoji':
                        stream = await self.http.get_twemoji(content)
                    else:
                        stream = (
                            await self.http.get_discord_emoji(content)
                            if self.render_discord_emoji else None
                        )

                    if not stream:
                        self.draw.text((x, y), content, *args, **kwargs)
                    else:
                        with Image.open(stream).convert("RGBA") as asset:
                            asset = asset.resize((width := int(emoji_size_factor * font.size), width), Image.ANTIALIAS)
                            box = x + emoji_position_offset[0], y + emoji_position_offset[1]
                            self.image.paste(asset, box, asset)

                x += width
            y += spacing + font.size



import typing
from PIL import ImageFont
from .helpers import get_nodes


__all__ = [
    'BaseRequester',
    'BasePilmoji'
]


class BasePilmoji:
    """
    The base emoji render that all renderes must inherit from.
    """
    @staticmethod
    def getsize(text: str, font=None, *, spacing=4,
                emoji_size_factor: float = 1,
                emoji_position_offset=(0, 0)) -> typing.Tuple[int, int]:
        """
        Gets the size of the given text.
        :param text: The text to use.
        :param font: The font to use.
        :param spacing: The line spacing, in pixels.
        :param emoji_size_factor: The emoji resize factor.
        :param emoji_position_offset: The emoji's position offset.
        :return: A tuple with values width, height.
        """
        if not font:
            font = ImageFont.load_default()

        x, y = 0, 0
        nodes = get_nodes(text.split('\n'))

        for line in nodes:
            this_x = 0
            for node in line:
                content = node['content']
                width, _ = font.getsize(content)
                if node['type'] != 'text':
                    width = int(emoji_size_factor * font.size)

                this_x += width
            y += spacing + font.size
            if this_x > x:
                x = this_x

        return x, y - spacing



class BaseRequester:
    """
    The base requester that all requesters must inherit from.
    """
    BASE_URL = 'https://twemoji.maxcdn.com/v/latest/72x72/'
    BASE_DISCORD_URL = 'https://cdn.discordapp.com/emojis/'
    BASE_MICROSOFT_URL = 'https://emojipedia-us.s3.dualstack.us-west-1.amazonaws.com/thumbs/120/microsoft/209/'



import re
from emoji.unicode_codes import EMOJI_UNICODE


ALL_EMOJIS = list(EMOJI_UNICODE['en'].values())

_TWEMOJI_EMOJI_REGEX = '|'.join(map(re.escape, ALL_EMOJIS))
EMOJI_REGEX = re.compile(f"(?P<twemoji>{_TWEMOJI_EMOJI_REGEX})|"+r'<a?:[a-zA-Z0-9_]{2,32}:(?P<discord>[0-9]{17,})>')



import aiohttp
import asyncio
import typing
import unicodedata
from io import BytesIO
from ..classes import BaseRequester


__all__ = [
    'AsyncRequester'
]


class AsyncRequester(BaseRequester):
    """
    Makes requests to fetch images using aiohttp.
    Use `Requester` for non-asynchronous requests.
    """
    def __init__(self, *, session: typing.Optional[aiohttp.ClientSession] = None, loop: typing.Optional[asyncio.AbstractEventLoop] = None, _microsoft=False):
        self.session: aiohttp.ClientSession = session or aiohttp.ClientSession(loop=loop)
        self.cache: typing.Dict[str, bytes] = {}
        self.loop: asyncio.AbstractEventLoop = loop
        self._microsoft = _microsoft

    async def _request(self, url) -> typing.Optional[BytesIO]:
        if url in self.cache:
            return BytesIO(self.cache[url])

        async with self.session.get(url) as response:
            if response.status == 200:
                stream = BytesIO(content := await response.read())
                self.cache[url] = content
                return stream

    async def get_twemoji(self, unicode: str) -> typing.Optional[BytesIO]:
        """
        Returns a stream of the given unicode emoji.
        :param unicode: The unicode emoji.
        :return: The bytes stream of that emoji.
        """
        hex_code = format(ord(unicode[0]), 'x')
        if not self._microsoft:
            url = self.BASE_URL + hex_code + '.png'
        else:
            name = unicodedata.name(unicode, hex_code)
            name = name.lower().replace(' ', '-')
            url = self.BASE_MICROSOFT_URL + name + f'_{hex_code}.png'
        return await self._request(url)

    async def get_discord_emoji(self, emoji_id: typing.Union[int, str]) -> typing.Optional[BytesIO]:
        """
        Returns a stream of the given Discord emoji.
        :param emoji_id: The emoji's ID.
        :return: A bytes stream of the emoji.
        """
        url = self.BASE_DISCORD_URL + str(emoji_id) + '.png?v=1'
        return await self._request(url)

    async def close(self):
        """
        Closes the aiohttp session.
        This will also delete the cache as it will become pointless.
        """
        await self.session.close()
        del self.session
        del self.cache

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()



from .constants import EMOJI_REGEX


def get_nodes(lines):
    mappings = []
    for line in lines:
        buffer = []
        for match in EMOJI_REGEX.finditer(line):
            if specific := match.group('twemoji'):
                buffer.append({
                    'type': 'twemoji',
                    'text': specific,
                    'span': match.span()
                })
            elif specific := match.group('discord'):
                buffer.append({
                    'type': 'discord',
                    'text': specific,
                    'span': match.span()
                })
        mappings.append(buffer)

    final = []
    for line, mapping in zip(lines, mappings):
        buffer = []
        if not len(mapping):
            buffer.append({
                'content': line,
                'type': 'text'
            })
        else:
            idx = 0
            for data in mapping:
                buffer.append({
                    'content': line[idx:data['span'][0]],
                    'type': 'text'
                })
                buffer.append({
                    'content': data['text'],
                    'type': data['type']
                })
                idx = data['span'][1]

            buffer.append({
                'content': line[idx:],
                'type': 'text'
            })
        final.append(buffer)

    return final



import setuptools

with open("README.md", "r", encoding='utf-8') as f:
    _description = f.read()

setuptools.setup(
    name="pilmoji_fixed",
    version="1.3.2",
    author="jay3332 (fix by stunnerr)",
    description="Pilmoji is a fast and reliable emoji renderer for PIL.",
    long_description=_description,
    long_description_content_type="text/markdown",
    url="https://github.com/stunnerr/pilmoji",
    packages=setuptools.find_packages(),
    install_requires=[
        "requests",
        "aiohttp",
        "pillow",
        "emoji",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
